/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*-
 * Copyright (c) 1996 Berkeley Software Design, Inc.
 * All rights reserved.
 * The Berkeley Software Design Inc. software License Agreement specifies
 * the terms and conditions for redistribution.
 *
 *      BSDI sysctl.c,v 2.14 1996/12/31 20:37:45 karels Exp
 */

/*
 * Copyright (c) 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright (c) 1993\n\
	The Regents of the University of California.  All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char sccsid[] = "@(#)sysctl.c	8.5 (Berkeley) 5/9/95";
#endif /* not lint */

#include <sys/param.h>
#include <sys/gmon.h>
#include <sys/mount.h>
#include <sys/stat.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <sys/protosw.h>
#include <sys/license.h>
#include <vm/vm_param.h>
#include <machine/cpu.h>

#include <net/if.h>

#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/icmp_var.h>
#include <netinet/igmp_var.h>
#include <netinet/ip_var.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/udp_var.h>

#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <netinet6/ip6_var.h>
#include <netinet6/tcp6.h>
#include <netinet6/udp6_var.h>
#ifdef IPSEC
#include <net/route.h>
#include <netinet6/ipsec.h>
#endif /* IPSEC */

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "license.h"
#include "extern.h"

struct ctlname topname[] = CTL_NAMES;
struct ctlname kernname[] = CTL_KERN_NAMES;
struct ctlname vmname[] = CTL_VM_NAMES;
struct ctlname netname[] = CTL_NET_NAMES;
struct ctlname hwname[] = CTL_HW_NAMES;
struct ctlname username[] = CTL_USER_NAMES;
struct ctlname debugname[CTL_DEBUG_MAXID];
struct ctlname *vfsname;
#ifdef CTL_MACHDEP_NAMES
struct ctlname machdepname[] = CTL_MACHDEP_NAMES;
#endif
char names[BUFSIZ];
int lastused;

struct list {
	struct	ctlname *list;
	int	size;
};
struct list toplist = { topname, CTL_MAXID };
struct list secondlevel[] = {
	{ 0, 0 },			/* CTL_UNSPEC */
	{ kernname, KERN_MAXID },	/* CTL_KERN */
	{ vmname, VM_MAXID },		/* CTL_VM */
	{ 0, 0 },			/* CTL_VFS */
	{ netname, NET_MAXID },		/* CTL_NET */
	{ 0, CTL_DEBUG_MAXID },		/* CTL_DEBUG */
	{ hwname, HW_MAXID },		/* CTL_HW */
#ifdef CTL_MACHDEP_NAMES
	{ machdepname, CPU_MAXID },	/* CTL_MACHDEP */
#else
	{ 0, 0 },			/* CTL_MACHDEP */
#endif
	{ username, USER_MAXID },	/* CTL_USER_NAMES */
};

int	flags, nflag, wflag;

/*
 * Variables requiring special processing.
 */
#define	CLOCK		0x00000001
#define	BOOTTIME	0x00000002
#define	CONSDEV		0x00000004
#define	SLOWHZ		0x00000008	/* convert seconds to/from PR_SLOWHZ */

#define	FLAG_A		0x01	/* The A Flag was given */
#define	FLAG_a		0x02	/* The a Flag was given */

void	debuginit __P((void));
int	findname __P((char *, char *, char **, struct list *));
void	listall __P((char *, struct list *));
int	parse __P((char *, int));
int	sysctl_inet __P((char *, char **, int *, int, int *, int *));
int	sysctl_inet6 __P((char *, char **, int *, int, int *, int *));
int	sysctl_route __P((char *, char **, int *, int, int *));
int	sysctl_socket __P((char *, char **, int *, int *));
void	usage __P((void));
void	vfsinit __P((void));

int
main(argc, argv)
	int argc;
	char *argv[];
{
	extern int optind;
	int ch, eval, lvl1;

	while ((ch = getopt(argc, argv, "Aanw")) != EOF) {
		switch (ch) {
		case 'A':
			flags |= FLAG_A;
			break;
		case 'a':
			flags |= FLAG_a;
			break;
		case 'n':
			nflag = 1;
			break;
		case 'w':
			wflag = 1;
			break;
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;

	if (argc == 0 && flags & (FLAG_A | FLAG_a)) {
		debuginit();
		vfsinit();
		for (lvl1 = 1; lvl1 < CTL_MAXID; lvl1++)
			listall(topname[lvl1].ctl_name, &secondlevel[lvl1]);
		exit(0);
	}
	if (argc == 0)
		usage();
	for (eval = 0; *argv != NULL; ++argv)
		if (parse(*argv, 1))
			eval = 1;
	exit(eval);
}

/*
 * List all variables known to the system.
 */
void
listall(prefix, lp)
	char *prefix;
	struct list *lp;
{
	int lvl2;
	char name[BUFSIZ];

	if (lp->list == 0)
		return;
	for (lvl2 = 0; lvl2 < lp->size; lvl2++) {
		if (lp->list[lvl2].ctl_name == 0)
			continue;
		(void)snprintf(name, sizeof(name),
		    "%s.%s", prefix, lp->list[lvl2].ctl_name);
		(void)parse(name, flags);
	}
}

/*
 * Parse a name into a MIB entry.  Lookup and print out the MIB entry if it
 * exists.  Set a new value if requested.
 */
int
parse(string, flags)
	char *string;
	int flags;
{
	license_t lic, nlic;
	struct list *lp;
	struct vfsconf vfc;
	size_t size;
	u_int intval;
	u_quad_t quadval;
	void *newval;
	int indx, len, newsize, special, state, type;
	int mib[CTL_MAXNAME];
	char *cp, *bufp, buf[BUFSIZ];

	newval = NULL;
	newsize = special = 0;

	bufp = buf;
	snprintf(buf, BUFSIZ, "%s", string);
	if ((cp = strchr(string, '=')) != NULL) {
		if (!wflag)
			errx(2, "must specify -w to set variables");
		*strchr(buf, '=') = '\0';
		*cp++ = '\0';
		while (isspace(*cp))
			cp++;
		newval = cp;
		newsize = strlen(cp);
	}
	if ((indx = findname(string, "top", &bufp, &toplist)) == -1)
		return (1);
	mib[0] = indx;
	if (indx == CTL_VFS)
		vfsinit();
	if (indx == CTL_DEBUG)
		debuginit();
	lp = &secondlevel[indx];
	if (lp->list == 0) {
		warnx("%s: class is not implemented", topname[indx].ctl_name);
		return (1);
	}
	if (bufp == NULL) {
		listall(topname[indx].ctl_name, lp);
		return (0);
	}
	if ((indx = findname(string, "second", &bufp, lp)) == -1)
		return (1);
	mib[1] = indx;
	type = lp->list[indx].ctl_type;
	len = 2;
	switch (mib[0]) {
	case CTL_KERN:
		switch (mib[1]) {
		case KERN_PROF:
			mib[2] = GPROF_STATE;
			size = sizeof(state);
			if (sysctl(mib, 3, &state, &size, NULL, 0) < 0) {
				if (!(flags & FLAG_A))
					return (0);
				if (!nflag)
					fprintf(stderr, "%s: ", string);
				warnx("this kernel not compiled for profiling");
				return (1);
			}
			if (!nflag)
				printf("%s: %s\n", string,
				    state == GMON_PROF_OFF ? "off" : "running");
			return (0);
		case KERN_VNODE:
		case KERN_FILES:
			if (!(flags & FLAG_A))
				return (0);
			warnx("use pstat to view %s information", string);
			return (1);
		case KERN_PROC:
			if (!(flags & FLAG_A))
				return (0);
			warnx("use ps to view %s information", string);
			return (1);
		case KERN_CLOCKRATE:
			special |= CLOCK;
			break;
		case KERN_BOOTTIME:
			special |= BOOTTIME;
			break;
		case KERN_SYSVIPC:
			if (!(flags & FLAG_A))
				return (0);
			warnx("use ipcs to view %s information", string);
			return (1);
		case KERN_LICENSE:
			size = sizeof(lic);
			if (newval) {
				if ((errno =
				    license_decode(newval, &nlic)) != 0) {
					if (!nflag)
						fprintf(stderr, "%s: ", string);
					warnx("invalid license: %s", newval);
					return (1);
				}
			}
			if (sysctl(mib, 2, &lic, &size,
			    newval ? &nlic : NULL, newval ? size : 0) < 0) {
				if (!(flags & FLAG_A))
					return (0);
				if (!nflag)
					fprintf(stderr, "%s: ", string);
				if (errno == EOPNOTSUPP)
					warnx("this kernel not compiled for licenses");
				else if (errno == EINVAL)
					warnx("invalid license: %s", newval);
				else
					warn(NULL);
				return (1);
			}
			if (!nflag)
				printf("%s: ", string);
			printf("%s", license_encode(&lic));
			if (newval)
				printf(" -> %s", newval);
			printf("\n");
			return (0);
		}
		break;
	case CTL_HW:
		break;
	case CTL_VM:
		switch (mib[1]) {
		case VM_LOADAVG: {
			double loads[3];

			getloadavg(loads, 3);
			if (!nflag)
				printf("%s: ", string);
			printf("%.2f %.2f %.2f\n",
			    loads[0], loads[1], loads[2]);
			return (0);
		}
		case VM_TOTAL:
		case VM_CNT:
		case VM_SWAPSTATS:
			if (!(flags & FLAG_A))
				return (0);
			warnx("Use vmstat or systat to view %s information",
			    string);
			return (1);
		default:
			break;
		}
		break;
	case CTL_NET:
		switch (mib[1]) {
		case PF_UNSPEC:		/* socket layer */
			len = sysctl_socket(string, &bufp, mib, &type);
			if (len >= 0)
				goto doit;
			return (0);
		case PF_INET:
			len = sysctl_inet(string,
			    &bufp, mib, flags & FLAG_A, &type, &special);
			if (len >= 0)
				goto doit;
			return (0);
		case PF_INET6:
			len = sysctl_inet6(string,
			    &bufp, mib, flags & FLAG_A, &type, &special);
			if (len >= 0)
				goto doit;
			return (0);
		case PF_ROUTE:
			len = sysctl_route(string,
			    &bufp, mib, flags & FLAG_A, &type);
			if (len >= 0)
				goto doit;
			return (0);
		case PF_LINK:
			len = sysctl_link(string,
			    &bufp, mib, flags & FLAG_A, &type);
			if (len >= 0)
				goto doit;
			return (0);
		}
		if (!(flags & FLAG_A))
			return (0);
		warnx("Use netstat to view %s information", string);
		return (1);
	case CTL_DEBUG:
		mib[2] = CTL_DEBUG_VALUE;
		len = 3;
		break;
	case CTL_MACHDEP:
#ifdef CPU_CONSDEV
		if (mib[1] == CPU_CONSDEV)
			special |= CONSDEV;
#endif
		break;
	case CTL_VFS:
		mib[3] = mib[1];
		mib[1] = VFS_GENERIC;
		mib[2] = VFS_CONF;
		len = 4;
		size = sizeof(vfc);
		if (sysctl(mib, 4, &vfc, &size, NULL, (size_t)0) < 0) {
			warn("vfs print");
			return (1);
		}
		if (flags == 0 && vfc.vfc_refcount == 0)
			return (0);
		if (!nflag)
			printf("%s has %d mounted instance%s\n",
			    string, vfc.vfc_refcount,
			    vfc.vfc_refcount != 1 ? "s" : "");
		else
			printf("%d\n", vfc.vfc_refcount);
		return (0);
	case CTL_USER:
		break;
	default:
		warnx("illegal top level value: %d", mib[0]);
		return (1);
	}
doit:
	if (bufp) {
		warnx("name %s in %s is unknown", bufp, string);
		return (1);
	}
	if (newval) {
		switch (type) {
		case CTLTYPE_INT:
			intval = strtoul(newval, &cp, 0);
			/*
			 * Check that value is not empty and
			 * has at most one suffix character.
			 */
			if (cp == newval || (cp[0] != '\0' && cp[1] != '\0')) {
				warnx("invalid integer value %s", newval);
				return (1);
			}
			switch (*cp) {
			case '\0':
				break;
			case 'k':
			case 'K':
				intval *= 1024;
				break;
			case 'm':
			case 'M':
				intval *= 1048576;
				break;
			default:
				warnx("invalid integer value %s", newval);
				return (1);
			}
			newval = &intval;
			newsize = sizeof(intval);
			break;
		case CTLTYPE_QUAD:
			quadval = strtouq(newval, &cp, 0);
			/*
			 * Check that value is not empty and
			 * has at most one suffix character.
			 */
			if (cp == newval || (cp[0] != '\0' && cp[1] != '\0')) {
				warnx("invalid integer value %s", newval);
				return (1);
			}
			switch (*cp) {
			case '\0':
				break;
			case 'k':
			case 'K':
				quadval *= 1024;
				break;
			case 'm':
			case 'M':
				quadval *= 1048576;
				break;
			default:
				warnx("invalid integer value %s", newval);
				return (1);
			}
			newval = &quadval;
			newsize = sizeof(quadval);
			break;
		}
	}
	size = BUFSIZ;
	/* for SLOWHZ, convert any new value from seconds to PR_SLOWHZ */
	if (special & SLOWHZ && newsize && *(int *) newval)
		*(int *)newval *= PR_SLOWHZ;
	if (sysctl(mib, len, buf, &size, newsize ? newval : 0, newsize) == -1) {
		if (!(flags & FLAG_A))
			return (0);
		switch (errno) {
		case EOPNOTSUPP:
			warnx("%s: value is not available", string);
			return (1);
		case ENOTDIR:
			warnx("%s: specification is incomplete", string);
			return (1);
		case ENOMEM:
			warnx("%s: type is unknown", string);
			return (1);
		default:
			warn("%s", string);
			return (1);
		}
	}
	if (special & CLOCK) {
		struct clockinfo *clkp = (struct clockinfo *)buf;

		if (!nflag)
			printf("%s: ", string);
		printf(
		    "hz = %d, tick = %d, profhz = %d, stathz = %d\n",
		    clkp->hz, clkp->tick, clkp->profhz, clkp->stathz);
		return (0);
	}
	if (special & BOOTTIME) {
		struct timeval *btp = (struct timeval *)buf;

		if (!nflag)
			printf("%s = %s", string, ctime(&btp->tv_sec));
		else
			printf("%ld\n", btp->tv_sec);
		return (0);
	}
	if (special & CONSDEV) {
		dev_t dev = *(dev_t *)buf;
		char *p;

		if (!nflag) {
			if ((p = devname(dev, S_IFCHR)) == NULL)
				p = "??";
			printf("%s = %s\n", string, p);
		} else
			printf("0x%lx\n", dev);
		return (0);
	}
	if (special & SLOWHZ) {
		/* convert values to seconds */
		*(int *)buf /= PR_SLOWHZ;
		if (newsize && *(int *) newval)
			*(int *)newval /= PR_SLOWHZ;
	}
	switch (type) {
	case CTLTYPE_INT:
		if (newsize == 0) {
			if (!nflag)
				printf("%s = ", string);
			printf("%d\n", *(int *)buf);
		} else {
			if (!nflag)
				printf("%s: %d -> ", string, *(int *)buf);
			printf("%d\n", *(int *)newval);
		}
		return (0);
	case CTLTYPE_STRING:
		if (newsize == 0) {
			if (!nflag)
				printf("%s = ", string);
			printf("%s\n", buf);
		} else {
			if (!nflag)
				printf("%s: %s -> ", string, buf);
			printf("%s\n", (char *)newval);
		}
		return (0);
	case CTLTYPE_QUAD:
		if (newsize == 0) {
			if (!nflag)
				printf("%s = ", string);
			printf("%qd\n", *(quad_t *)buf);
		} else {
			if (!nflag)
				printf("%s: %qd -> ", string, *(quad_t *)buf);
			printf("%qd\n", *(quad_t *)newval);
		}
		return (0);
	case CTLTYPE_STRUCT:
		if (flags == 0) {
			warnx("%s: unknown structure returned", string);
			return (1);
		}
		return (0);
	default:
	case CTLTYPE_NODE:
		if (flags == 0) {
			warnx("%s: unknown type returned", string);
			return (1);
		}
		return (0);
	}
	/* NOTREACHED */
}

/*
 * Initialize the set of debugging names
 */
void
debuginit()
{
	size_t size;
	int mib[3], loc, i;

	if (secondlevel[CTL_DEBUG].list != 0)
		return;
	secondlevel[CTL_DEBUG].list = debugname;
	mib[0] = CTL_DEBUG;
	mib[2] = CTL_DEBUG_NAME;
	for (loc = lastused, i = 0; i < CTL_DEBUG_MAXID; i++) {
		mib[1] = i;
		size = BUFSIZ - loc;
		if (sysctl(mib, 3, &names[loc], &size, NULL, 0) == -1)
			continue;
		debugname[i].ctl_name = &names[loc];
		debugname[i].ctl_type = CTLTYPE_INT;
		loc += size;
	}
	lastused = loc;
}

/*
 * Initialize the set of filesystem names
 */
void
vfsinit()
{
	struct vfsconf vfc;
	size_t buflen;
	int mib[4], maxtypenum, cnt, loc, size;

	if (secondlevel[CTL_VFS].list != 0)
		return;
	mib[0] = CTL_VFS;
	mib[1] = VFS_GENERIC;
	mib[2] = VFS_MAXTYPENUM;
	buflen = 4;
	if (sysctl(mib, 3, &maxtypenum, &buflen, NULL, (size_t)0) < 0)
		return;
	if ((vfsname = malloc(maxtypenum * sizeof(*vfsname))) == 0)
		return;
	memset(vfsname, 0, maxtypenum * sizeof(*vfsname));
	mib[2] = VFS_CONF;
	buflen = sizeof(vfc);
	for (loc = lastused, cnt = 0; cnt < maxtypenum; cnt++) {
		mib[3] = cnt;
		if (sysctl(mib, 4, &vfc, &buflen, NULL, (size_t)0) < 0) {
			if (errno == EOPNOTSUPP)
				continue;
			warn("vfsinit");
			free(vfsname);
			return;
		}
		strcat(&names[loc], vfc.vfc_name);
		vfsname[cnt].ctl_name = &names[loc];
		vfsname[cnt].ctl_type = CTLTYPE_INT;
		size = strlen(vfc.vfc_name) + 1;
		loc += size;
	}
	lastused = loc;
	secondlevel[CTL_VFS].list = vfsname;
	secondlevel[CTL_VFS].size = maxtypenum;
	return;
}

struct ctlname sockname[] = SOCTL_NAMES;
struct list socklist = { sockname, SOCTL_MAXID };

/*
 * handle socket requests
 */
/* ARGSUSED4 */
int
sysctl_socket(string, bufpp, mib, typep)
	char *string;
	char **bufpp;
	int mib[];
	int *typep;
{
	int indx;

	if (*bufpp == NULL) {
		listall(string, &socklist);
		return (-1);
	}
	if ((indx = findname(string, "third", bufpp, &socklist)) == -1)
		return (-1);
	mib[2] = indx;
	*typep = sockname[indx].ctl_type;
	return (3);
}

struct ctlname inetname[] = CTL_IPPROTO_NAMES;
struct ctlname ipname[] = IPCTL_NAMES;
struct ctlname icmpname[] = ICMPCTL_NAMES;
struct ctlname igmpname[] = IGMPCTL_NAMES;
struct ctlname tcpname[] = TCPCTL_NAMES;
struct ctlname udpname[] = UDPCTL_NAMES;
struct ctlname ipsecname[] = IPSECCTL_NAMES;
struct list inetlist = { inetname, IPPROTO_MAXID };
struct list inetvars[] = {
/*0*/	{ ipname, IPCTL_MAXID },	/* ip */
	{ icmpname, ICMPCTL_MAXID },	/* icmp */
	{ igmpname, IGMPCTL_MAXID },	/* igmp */
	{ 0, 0 },			/* ggmp */
	{ 0, 0 },
/*5*/	{ 0, 0 },
	{ tcpname, TCPCTL_MAXID },	/* tcp */
	{ 0, 0 },
	{ 0, 0 },			/* egp */
	{ 0, 0 },
/*10*/	{ 0, 0 },
	{ 0, 0 },
	{ 0, 0 },			/* pup */
	{ 0, 0 },
	{ 0, 0 },
/*15*/	{ 0, 0 },
	{ 0, 0 },
	{ udpname, UDPCTL_MAXID },	/* udp */
	{ 0, 0 },
	{ 0, 0 },
/*20*/	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*25*/	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*30*/	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*35*/	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*40*/	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*45*/	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*50*/	{ ipsecname, IPSECCTL_MAXID },	/* ipsec */
};

/*
 * handle internet requests
 */
int
sysctl_inet(string, bufpp, mib, flags, typep, specialp)
	char *string;
	char **bufpp;
	int mib[];
	int flags;
	int *typep;
	int *specialp;
{
	struct list *lp;
	int indx;

	if (*bufpp == NULL) {
		listall(string, &inetlist);
		return (-1);
	}
	if ((indx = findname(string, "third", bufpp, &inetlist)) == -1)
		return (-1);
	mib[2] = indx;
	if (indx <= IPPROTO_ESP && inetvars[indx].list != NULL)
		lp = &inetvars[indx];
	else if (!flags)
		return (-1);
	else {
		warnx("%s: no variables defined for this protocol", string);
		return (-1);
	}
	if (*bufpp == NULL) {
		listall(string, lp);
		return (-1);
	}
	if ((indx = findname(string, "fourth", bufpp, lp)) == -1)
		return (-1);
	mib[3] = indx;
	*typep = lp->list[indx].ctl_type;
	switch (mib[2]) {
	case IPPROTO_TCP:
		switch (mib[3]) {
		case TCPCTL_KEEPIDLE:
		case TCPCTL_KEEPINTVL:
		case TCPCTL_MAXPERSISTIDLE:
		case TCPCTL_CONNTIMEO:
			*specialp |= SLOWHZ;
			break;
		default:
			break;
		}
		break;
	default:
		break;
	}
	return (4);
}

struct ctlname inet6name[] = CTL_IPV6PROTO_NAMES;
struct ctlname ip6name[] = IPV6CTL_NAMES;
struct ctlname icmp6name[] = ICMPV6CTL_NAMES;
struct ctlname tcp6name[] = TCP6CTL_NAMES;
struct ctlname udp6name[] = UDP6CTL_NAMES;
struct ctlname ipsec6name[] = IPSEC6CTL_NAMES;
struct list inet6list = { inet6name, IPV6PROTO_MAXID };
struct list inet6vars[] = {
/*0*/	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*5*/	{ 0, 0 },
	{ tcp6name, TCP6CTL_MAXID },	/* tcp */
	{ 0, 0 },
	{ 0, 0 },
	{ 0, 0 },
/*10*/	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*15*/	{ 0, 0 },
	{ 0, 0 },
	{ udp6name, UDP6CTL_MAXID },	/* udp */
	{ 0, 0 },
	{ 0, 0 },
/*20*/	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*25*/	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*30*/	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*35*/	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*40*/	{ 0, 0 },
	{ ip6name, IPV6CTL_MAXID },	/* ip6 */
	{ 0, 0 },
	{ 0, 0 },
	{ 0, 0 },
/*45*/	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*50*/	{ ipsec6name, IPSECCTL_MAXID },	/* ipsec6 */
	{ 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 },
/*55*/	{ 0, 0 },
	{ 0, 0 },
	{ 0, 0 },
	{ icmp6name, ICMPV6CTL_MAXID },	/* icmp6 */
};

/*
 * handle internet requests
 */
int
sysctl_inet6(string, bufpp, mib, flags, typep, specialp)
	char *string;
	char **bufpp;
	int mib[];
	int flags;
	int *typep;
	int *specialp;
{
	struct list *lp;
	int indx;

	if (*bufpp == NULL) {
		listall(string, &inet6list);
		return (-1);
	}
	if ((indx = findname(string, "third", bufpp, &inet6list)) == -1)
		return (-1);
	mib[2] = indx;
	if (indx <= IPPROTO_ICMPV6 && inet6vars[indx].list != NULL)
		lp = &inet6vars[indx];
	else if (!flags)
		return (-1);
	else {
		warnx("%s: no variables defined for this protocol", string);
		return (-1);
	}
	if (*bufpp == NULL) {
		listall(string, lp);
		return (-1);
	}
	if ((indx = findname(string, "fourth", bufpp, lp)) == -1)
		return (-1);
	mib[3] = indx;
	*typep = lp->list[indx].ctl_type;
	switch (mib[2]) {
	case IPPROTO_TCP:
		switch (mib[3]) {
		case TCP6CTL_KEEPIDLE:
		case TCP6CTL_KEEPINTVL:
		case TCP6CTL_MAXPERSISTIDLE:
		case TCP6CTL_CONNTIMEO:
			*specialp |= SLOWHZ;
			break;
		default:
			break;
		}
		break;
	default:
		break;
	}
	return (4);
}




struct ctlname routename[] = CTL_NET_ROUTE_NAMES;
struct list routelist = { routename, NET_ROUTE_MAXID };

struct ctlname arpname[] = ARPCTL_NAMES;
struct list arplist = { arpname, ARPCTL_MAXID };

/*
 * handle route requests
 */
int
sysctl_route(string, bufpp, mib, flags, typep)
	char *string;
	char **bufpp;
	int mib[];
	int flags;
	int *typep;
{
	int indx;

	if (*bufpp == NULL) {
		listall(string, &routelist);
		return (-1);
	}

	if ((indx = findname(string, "third", bufpp, &routelist)) == -1)
		return (-1);

	if (indx == NET_ROUTE_TABLE) {
		if (flags & FLAG_A)
			warnx("Use netstat to view %s information", string);
		return(-1);
	}

	mib[2] = indx;
	if (indx != NET_ROUTE_ARP) {
		*typep = routename[indx].ctl_type;
		return (3);
	}

	if (*bufpp == NULL) {
		listall(string, &arplist);
		return (-1);
	}

	if ((indx = findname(string, "fourth", bufpp, &arplist)) == -1)
		return (-1);

	mib[3] = indx;
	*typep = arpname[indx].ctl_type;
	return (4);
}

/* XXX these are all placeholders; should be moved into proper headers */
#define CTL_LINK_MAXID	1
struct ctlname linkname[] = { "generic", CTLTYPE_NODE };
struct list linklist = { linkname, CTL_LINK_MAXID };

struct ctlname linktypename[] = {
	{ 0, 0 },			/* 0x00 */
	{ 0, 0 },			/* 0x01 */
	{ 0, 0 },			/* 0x02 */
	{ 0, 0 },			/* 0x03 */
	{ 0, 0 },			/* 0x04 */
	{ 0, 0 },			/* 0x05 */
	{ 0, 0 },			/* 0x06 */
	{ 0, 0 },			/* 0x07 */
	{ 0, 0 },			/* 0x08 */
	{ 0, 0 },			/* 0x09 */
	{ 0, 0 },			/* 0x0a */
	{ 0, 0 },			/* 0x0b */
	{ 0, 0 },			/* 0x0c */
	{ 0, 0 },			/* 0x0d */
	{ 0, 0 },			/* 0x0e */
	{ 0, 0 },			/* 0x0f */
	{ 0, 0 },			/* 0x10 */
	{ 0, 0 },			/* 0x11 */
	{ 0, 0 },			/* 0x10 */
	{ 0, 0 },			/* 0x13 */
	{ 0, 0 },			/* 0x14 */
	{ 0, 0 },			/* 0x15 */
	{ 0, 0 },			/* 0x16 */
	{ "ppp", CTLTYPE_NODE },	/* 0x17 */
	{ 0, 0 },			/* 0x18 */
	{ 0, 0 },			/* 0x19 */
	{ 0, 0 },			/* 0x1a */
	{ 0, 0 },			/* 0x1b */
	{ 0, 0 },			/* 0x1c */
	{ 0, 0 },			/* 0x1d */
	{ 0, 0 },			/* 0x1e */
	{ 0, 0 },			/* 0x1f */
	{ 0, 0 },			/* 0x20 */
	{ 0, 0 },			/* 0x21 */
	{ 0, 0 },			/* 0x22 */
	{ 0, 0 },			/* 0x23 */
	{ 0, 0 },			/* 0x24 */
	{ 0, 0 },			/* 0x25 */
	{ 0, 0 },			/* 0x26 */
	{ 0, 0 },			/* 0x27 */
	{ 0, 0 },			/* 0x28 */
	{ 0, 0 },			/* 0x29 */
	{ 0, 0 },			/* 0x2a */
	{ 0, 0 },			/* 0x2b */
	{ 0, 0 },			/* 0x2c */
	{ 0, 0 },			/* 0x2d */
	{ 0, 0 },			/* 0x2e */
	{ 0, 0 },			/* 0x2f */
	{ 0, 0 },			/* 0x30 */
	{ 0, 0 },			/* 0x31 */
	{ 0, 0 },			/* 0x32 */
	{ 0, 0 },			/* 0x33 */
	{ 0, 0 },			/* 0x34 */
	{ 0, 0 },			/* 0x35 */
	{ "pif", CTLTYPE_NODE },	/* 0x36 */
};
struct list linktypelist = { linktypename,
    sizeof(linktypename) / sizeof(linktypename[0]) };

#define CTL_LINKPROTO_MAXID	1
struct ctlname linkprotoname[] = { "numif", CTLTYPE_INT };
struct list linkvars[] = {
	{ 0, 0 },			/* 0x00 */
	{ 0, 0 },			/* 0x01 */
	{ 0, 0 },			/* 0x02 */
	{ 0, 0 },			/* 0x03 */
	{ 0, 0 },			/* 0x04 */
	{ 0, 0 },			/* 0x05 */
	{ 0, 0 },			/* 0x06 */
	{ 0, 0 },			/* 0x07 */
	{ 0, 0 },			/* 0x08 */
	{ 0, 0 },			/* 0x09 */
	{ 0, 0 },			/* 0x0a */
	{ 0, 0 },			/* 0x0b */
	{ 0, 0 },			/* 0x0c */
	{ 0, 0 },			/* 0x0d */
	{ 0, 0 },			/* 0x0e */
	{ 0, 0 },			/* 0x0f */
	{ 0, 0 },			/* 0x10 */
	{ 0, 0 },			/* 0x11 */
	{ 0, 0 },			/* 0x10 */
	{ 0, 0 },			/* 0x13 */
	{ 0, 0 },			/* 0x14 */
	{ 0, 0 },			/* 0x15 */
	{ 0, 0 },			/* 0x16 */
	{ linkprotoname, CTL_LINKPROTO_MAXID },	/* 0x17 */
	{ 0, 0 },			/* 0x18 */
	{ 0, 0 },			/* 0x19 */
	{ 0, 0 },			/* 0x1a */
	{ 0, 0 },			/* 0x1b */
	{ 0, 0 },			/* 0x1c */
	{ 0, 0 },			/* 0x1d */
	{ 0, 0 },			/* 0x1e */
	{ 0, 0 },			/* 0x1f */
	{ 0, 0 },			/* 0x20 */
	{ 0, 0 },			/* 0x21 */
	{ 0, 0 },			/* 0x22 */
	{ 0, 0 },			/* 0x23 */
	{ 0, 0 },			/* 0x24 */
	{ 0, 0 },			/* 0x25 */
	{ 0, 0 },			/* 0x26 */
	{ 0, 0 },			/* 0x27 */
	{ 0, 0 },			/* 0x28 */
	{ 0, 0 },			/* 0x29 */
	{ 0, 0 },			/* 0x2a */
	{ 0, 0 },			/* 0x2b */
	{ 0, 0 },			/* 0x2c */
	{ 0, 0 },			/* 0x2d */
	{ 0, 0 },			/* 0x2e */
	{ 0, 0 },			/* 0x2f */
	{ 0, 0 },			/* 0x30 */
	{ 0, 0 },			/* 0x31 */
	{ 0, 0 },			/* 0x32 */
	{ 0, 0 },			/* 0x33 */
	{ 0, 0 },			/* 0x34 */
	{ 0, 0 },			/* 0x35 */
	{ linkprotoname, CTL_LINKPROTO_MAXID },	/* 0x36 */
};

/*
 * handle link-level net requests
 * XXX this is a placeholder
 * For now, just handle
 *	net.link_layer.generic.{ppp,pif}.numif
 */
int
sysctl_link(string, bufpp, mib, flags, typep, specialp)
	char *string;
	char **bufpp;
	int mib[];
	int flags;
	int *typep;
	int *specialp;
{
	struct list *lp;
	int indx;

	if (*bufpp == NULL) {
		listall(string, &linklist);
		return (-1);
	}
	/*
	 * Maybe we shouldn't use findname; later the name will be
	 * "generic" or an interface name or index.  At least for now,
	 * this will make error-checking the same as elsewhere.
	 */
	if ((indx = findname(string, "third", bufpp, &linklist)) == -1)
		return (-1);
	mib[2] = indx;
	/* yep, it's "generic" */
	if (indx == 0)	/* generic */
		lp = &linktypelist;
#ifdef notyet
	else if (/* check for interface index by name, e.g. ppp0 => link#3 */)
		return (sysctl_interface(...));
#endif
	else if (!flags)
		return (-1);
	else {
		warnx("%s: no variables defined for this interface", string);
		return (-1);
	}

	if (*bufpp == NULL) {
		listall(string, lp);
		return (-1);
	}
	/* look up I/F type */
	if ((indx = findname(string, "fourth", bufpp, lp)) == -1)
		return (-1);
	mib[3] = indx;
	if (indx <= sizeof(linktypename) / sizeof(linktypename[0]) &&
	    linkvars[indx].list != NULL)
		lp = &linkvars[indx];
	else if (!flags)
		return (-1);
	else {
		warnx("%s: no variables defined for this interface type",
		    string);
		return (-1);
	}

	if (*bufpp == NULL) {
		listall(string, lp);
		return (-1);
	}
	/* look up variable name */
	if ((indx = findname(string, "fifth", bufpp, lp)) == -1)
		return (-1);
	mib[4] = indx;
	*typep = lp->list[indx].ctl_type;
	return (5);
}

/*
 * Scan a list of names searching for a particular name.
 */
int
findname(string, level, bufp, namelist)
	char *string;
	char *level;
	char **bufp;
	struct list *namelist;
{
	char *name;
	int i;

	if (namelist->list == 0 || (name = strsep(bufp, ".")) == NULL) {
		warnx("%s: incomplete specification", string);
		return (-1);
	}
	for (i = 0; i < namelist->size; i++)
		if (namelist->list[i].ctl_name != NULL &&
		    strcmp(name, namelist->list[i].ctl_name) == 0)
			break;
	if (i == namelist->size) {
		warnx("%s level name %s in %s is invalid", level, name, string);
		return (-1);
	}
	return (i);
}

void
usage()
{

	(void)fprintf(stderr, "usage:\t%s\n\t%s\n\t%s\n\t%s\n",
	    "sysctl [-n] variable ...",
	    "sysctl [-n] -w variable=value ...",
	    "sysctl [-n] -a",
	    "sysctl [-n] -A");
	exit(1);
}
