/*
 * Copyright (C) 1997 and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <errno.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "faithd.h"

char tcpbuf[MSS];

static void
relay(int s_rcv, int s_snd, const char *service)
{
	int err, atmark;
	ssize_t n;

	err = ioctl(s_rcv, SIOCATMARK, &atmark);
	if (err != -1 && atmark == 1) {
		n = read(s_rcv, tcpbuf, 1);
		if (n == 1)
			send(s_snd, tcpbuf, 1, MSG_OOB);
		return;
	}

	n = read(s_rcv, tcpbuf, sizeof(tcpbuf));
			
	switch (n) {
	case -1:
		exit_failure(ERRSTR);
	case 0:
		close(s_rcv);
		close(s_snd);
		exit_success("terminating %s relay", service);
		/* never reached */
	default:
		write(s_snd, tcpbuf, n);
#ifdef DEBUG
		if (n >= MSS)
			n = MSS - 1;
		tcpbuf[n] = '\0';
		if (s_rcv == s_src) 
			syslog(LOG_DEBUG, "from client: %s", tcpbuf);
		else 
			syslog(LOG_DEBUG, "from server: %s", tcpbuf);
#endif /* DEBUG */
	}
}

void
tcp_relay(int s_src, int s_dst, const char *service)
{
	fd_set readfds;
	int err;

	syslog(LOG_INFO, "starting %s relay", service);

	FD_ZERO(&readfds);

	for (;;) {
		FD_SET(s_src, &readfds);
		FD_SET(s_dst, &readfds);
		err = select(256, &readfds, NULL, NULL, NULL);
		if (err == -1)
			exit_failure("select: %s", ERRSTR);

		if (FD_ISSET(s_src, &readfds))
			relay(s_src, s_dst, service);
		if (FD_ISSET(s_dst, &readfds))
			relay(s_dst, s_src, service);
	}
	/* never reached */
}
