/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/socket.h>
#ifdef ADVAPI
#include <sys/uio.h>
#endif
#include <sys/ioctl.h>

#include <net/if.h>
#include <netinet/in.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <unistd.h>
#include <stdio.h>
#include <err.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <kvm.h>
#include <nlist.h>
#include <sys/sysctl.h>
#include <netinet6/in6.h>
#include <netinet6/in6_var.h>
#include <netinet6/nd6.h>

#define ALLROUTER "ff02::2"

void usage __P((void));
int sendit __P((int, char *, int, struct sockaddr_in6 *, int));
int interface_up __P((char *));
static int getifa __P((char *, struct in6_ifaddr *));
static int getdadcount __P((void));

static int verbose = 0;
#define vmsg(x)		{ if (verbose) fprintf x; }

extern char *optarg;
extern int optind;

int
main(argc, argv)
	int argc;
	char *argv[];
{
	struct sockaddr_in6 to;
	struct icmp6_hdr *icp;
	u_int hlim = 255;
	u_char outpack[sizeof(struct icmp6_hdr)];
	u_short index;
	int s, i, cc = sizeof(struct icmp6_hdr);
	int opt;
	fd_set fdset;
	struct timeval timeout;
	int rtsol_retry = MAX_RTR_SOLICITATIONS;
	
	while ((opt = getopt(argc, argv, "v")) != EOF) {
		switch (opt) {
		case 'v':
			verbose++;
			break;
		default:
			usage();
			/*NOTREACHED*/
		}
	}
	argc -= optind;
	argv += optind;
	if (argc != 1) {
		usage();
		/*NOTREACHED*/
	}

	index = (u_short)if_nametoindex(argv[0]);
	if (index == 0)
		usage();

	if (interface_up(argv[0]) < 0)
		err(1, "interface not ready");

	memset(&to, 0, sizeof(struct sockaddr_in6));
	if (inet_pton(AF_INET6, ALLROUTER, &to.sin6_addr.s6_addr) < 0)
		exit(1);
	to.sin6_len = sizeof(struct sockaddr_in6);
	to.sin6_family = AF_INET6;
	
	if ((s = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0)
		err(1, "socket");

	if (setsockopt(s, IPPROTO_IPV6, IPV6_MULTICAST_HOPS, &hlim,
			sizeof(hlim)) == -1)
		err(1, "setsockopt(IPV6_MULTICAST_HOPS)");
#ifdef ICMP6_FILTER
    {
	struct icmp6_filter filt;
	ICMP6_FILTER_SETBLOCKALL(&filt);
	ICMP6_FILTER_SETPASS(ND_ROUTER_ADVERT, &filt);
	if (setsockopt(s, IPPROTO_ICMPV6, ICMP6_FILTER, &filt,
			sizeof(filt)) == -1)
		err(1, "setsockopt(ICMP6_FILTER)");
    }
#endif
	
	/* fill ICMP header */
	icp = (struct icmp6_hdr *)outpack;
	icp->icmp6_type = ND_ROUTER_SOLICIT;
	icp->icmp6_code = 0;
	icp->icmp6_cksum = 0;

	vmsg((stderr, "sending RS for %d time%s:\n",
		rtsol_retry, (1 < rtsol_retry) ? "s" : ""));

retry:
	vmsg((stderr, "%d \r", rtsol_retry));

	i = sendit(s, (char *)outpack, cc, &to, index);
	if (i < 0) {
		perror("rtsol: solicitation");
		exit(1);
	} else if (i != cc) {
		fprintf(stderr, "rtsol: short write (wrote %d of be %d)\n",
			i, cc);
	}
	FD_ZERO(&fdset);
	FD_SET(s, &fdset);
	timeout.tv_sec = 1;
	timeout.tv_usec = 0;
	if (select(s + 1, &fdset, NULL, NULL, &timeout) < 1) {
		if (0 < --rtsol_retry)
			goto retry;
	} else {
		vmsg((stderr, "got a RA packet\n"));
	}

	close(s);
	exit(0);
}

void
usage()
{
	(void)fprintf(stderr, "usage: rtsol [-v] ifname\n");
	exit(1);
}

int
sendit(s, p, len, sock, ifindex)
	int s;
	char *p;
	int len;
	struct sockaddr_in6 *sock;
	int ifindex;
{
#ifndef ADVAPI
	sock->sin6_addr.u6_addr.u6_addr16[1] = htons(ifindex);
	return sendto(s, p, len, 0, (struct sockaddr *)sock, sock->sin6_len);
#else
	struct msghdr m;
	struct cmsghdr *cm;
	struct iovec iov[2];
	u_char cmsgbuf[256];
	struct in6_pktinfo *pi;

	m.msg_name = (caddr_t)sock;
	m.msg_namelen = sock->sin6_len;
	iov[0].iov_base = p;
	iov[0].iov_len = len;
	m.msg_iov = iov;
	m.msg_iovlen = 1;
	if (!ifindex) {
		m.msg_control = NULL;
		m.msg_controllen = 0;
	} else {
		memset(cmsgbuf, 0, sizeof(cmsgbuf));
		cm = (struct cmsghdr *)cmsgbuf;
		m.msg_control = (caddr_t)cm;
		m.msg_controllen = CMSG_SPACE(sizeof(struct in6_pktinfo));

		cm->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
		cm->cmsg_level = IPPROTO_IPV6;
		cm->cmsg_type = IPV6_PKTINFO;
		pi = (struct in6_pktinfo *)CMSG_DATA(cm);
		memset(&pi->ipi6_addr, 0, sizeof(pi->ipi6_addr));	/*XXX*/
		pi->ipi6_ifindex = ifindex;
	}
	return sendmsg(s, &m, 0);
#endif
}

int
interface_up(name)
	char *name;
{
	int s;
	struct ifreq ifr;
	struct in6_ifaddr ifa;
	int retry;
	int wasup;

	if ((s = socket(AF_INET6, SOCK_DGRAM, 0)) < 0) {
		perror("socket");
		exit(1);
	}
	strncpy(ifr.ifr_name, name, sizeof(ifr.ifr_name));

	if (ioctl(s, SIOCGIFFLAGS, (caddr_t)&ifr) < 0)
		perror("ioctl(SIOCGIFFLAGS)");
	if (ifr.ifr_flags & IFF_UP)
		wasup = 1;
	else {
		wasup = 0;
		ifr.ifr_flags |= IFF_UP;
		if (ioctl(s, SIOCSIFFLAGS, (caddr_t)&ifr) < 0)
			perror("ioctl(SIOCSIFFLAGS)");
	}

	retry = getdadcount();
	if (retry < 0) {
		retry = 10;
		vmsg((stderr, "invalid dad count; setting retry to %d\n",
			retry));
	} else {
		vmsg((stderr, "dad count is %d\n", retry));
		if (!wasup)
			retry += MAX_RTR_SOLICITATIONS;
		retry += 2;	/* to make sure */
	}
	vmsg((stderr, "checking if %s is ready...\n", name));
	while (0 < retry--) {
		if (getifa(name, &ifa) < 0) {
			vmsg((stderr, "getifa() failed, anyway I'll try\n"));
			return 0;
		}

		if (verbose) {
			fprintf(stderr, "%s:", name);
			if (!(ifa.ia6_flags & IN6_IFF_NOTREADY))
				fprintf(stderr, " ready\n");
			else {
				if (ifa.ia6_flags & IN6_IFF_TENTATIVE)
					fprintf(stderr, " tentative");
				if (ifa.ia6_flags & IN6_IFF_DUPLICATED)
					fprintf(stderr, " duplicatd");
				fprintf(stderr, "\n");
			}
		}

		if (ifa.ia6_flags & IN6_IFF_DUPLICATED)
			break;
		if (!(ifa.ia6_flags & IN6_IFF_NOTREADY))
			return 0;

		sleep(1);
	}
	fprintf(stderr, "DAD for if %s is in strange state\n", name);
	return -1;
}

/*------------------------------------------------------------*/

static struct nlist nl[] = {
#define N_IFNET	0
	{ "_ifnet" },
	{ "" },
};

#define KREAD(x, y, z) { \
	if (kvm_read(kvmd, (u_long)x, (void *)y, sizeof(z)) != sizeof(z)) { \
		perror("kvm_read");					\
		exit(1);						\
	}								\
   }

static int
getifa(name, ifap)
	char *name;
	struct in6_ifaddr *ifap;
{
	u_short index;
	kvm_t *kvmd;
	char buf[BUFSIZ];
	struct ifnet *ifp;
	struct ifnet ifnet;
	struct in6_ifaddr *ifa;

	if (!ifap)
		exit(1);

	index = (u_short)if_nametoindex(name);
	if (index == 0) {
		perror("if_nametoindex");
		return -1;
	}
	if ((kvmd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, buf)) == NULL) {
		perror("kvm_openfiles");
		return -1;
	}
	if (kvm_nlist(kvmd, nl) < 0) {
		perror("kvm_nlist");
		return -1;
	}
	if (nl[N_IFNET].n_value == 0) {
		fprintf(stderr, "symbol \"%s\" not found\n",
			nl[N_IFNET].n_name);
		return -1;
	}

	KREAD(nl[N_IFNET].n_value, &ifp, struct ifnet *);
	while (ifp) {
		KREAD(ifp, &ifnet, struct ifnet);
		if (ifnet.if_index == index)
			break;
#ifndef __NetBSD__
		ifp = ifnet.if_next;
#else
		ifp = ifnet.if_list.tqe_next;
#endif
	}
	if (!ifp) {
		fprintf(stderr, "interface \"%s\" not found\n", name);
		return -1;
	}

#ifndef __NetBSD__
	ifa = (struct in6_ifaddr *)ifnet.if_addrlist;
#else
	ifa = (struct in6_ifaddr *)ifnet.if_addrlist.tqh_first;
#endif
	while (ifa) {
		KREAD(ifa, ifap, *ifap);
		if (ifap->ia_addr.sin6_family == AF_INET6
		 && IN6_IS_ADDR_LINKLOCAL(&ifap->ia_addr.sin6_addr)) {
			return 0;
		}

#ifndef __NetBSD__
		ifa = (struct in6_ifaddr *)(((struct ifaddr *)ifap)->ifa_next);
#else
		ifa = (struct in6_ifaddr *)
			(((struct ifaddr *)ifap)->ifa_list.tqe_next);
#endif
	}
	fprintf(stderr, "no IPv6 link-local address for interface \"%s\"\n",
		name);
	return -1;
}

static int
getdadcount()
{
	int mib[] = { CTL_NET, PF_INET6, IPPROTO_IPV6, IPV6CTL_DAD_COUNT };
	int value;
	size_t size;

	size = sizeof(value);
	if (sysctl(mib, sizeof(mib)/sizeof(mib[0]), &value, &size, NULL, 0) < 0)
		return -1;
	else
		return value;
}
