/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
/* KAME $Id: setkey.c,v 1.1.4.1.2.5.2.21.2.9.2.1 1998/09/22 13:18:18 itojun Exp $ */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <net/route.h>
#include <netinet/in.h>
#include <netinet6/in6.h>
#include <netkey/keyv2.h>
#include <netkey/keydb.h>
#include <netkey/key_debug.h>

#if defined(PDEBUG)
#include <sys/un.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>

void sendkeyshort __P((u_int));
int sendkeymsg __P((void));
int postproc __P((int, caddr_t));
void dump_msg __P((struct sadb_msg *, int));
u_char *_str_addr __P((caddr_t, u_int, u_int, u_int));
u_char *_str_time __P((time_t t));

int f_debug = 0;
int f_verbose = 0;
int f_command = 0;
int f_script = 0;
int f_cmddump = 0;
int f_cmdflush = 0;
int f_dumpscript = 0;
int f_policy = 0;
char *pname;

u_char m_buf[BUFSIZ];
u_int m_len;

extern int lineno;

void Usage();

extern int parse(FILE **);

void
Usage()
{
	printf("Usage:\t%s [-c|-f (file)] [-dv]\n", pname);
	printf("\t%s [[-P] -D|-F] [-dv]\n", pname);
	exit(0);
}

int
main(ac, av)
	int ac;
	char **av;
{
	FILE *fp = stdin;
	int c;

	pname = *av;

	if (ac == 1) Usage();

	while ((c = getopt(ac, av, "cf:hdvDFP")) != EOF) {
		switch (c) {
		case 'c':
			f_script = 1;
			fp = stdin;
			break;
		case 'f':
			f_script = 1;
			if ((fp = fopen(optarg, "r")) == NULL) {
				perror("fopen");
				exit(-1);
			}
			break;
		case 'h':
			Usage();
			break;
		case 'd':
			f_debug = 1;
			break;
		case 'v':
			f_verbose = 1;
			break;
		case 'D':
			f_cmddump = 1;
			break;
		case 'F':
			f_cmdflush = 1;
			break;
		case 'P':
			/* XXX */
			f_policy = 1;
			break;
		default:
			Usage();
		}
	}

	if (f_cmdflush)
		sendkeyshort(f_policy ? SADB_X_SPDFLUSH: SADB_FLUSH);
	else
	if (f_cmddump)
		sendkeyshort(f_policy ? SADB_X_SPDDUMP: SADB_DUMP);
	else
	if (f_script)
		parse(&fp);
	else
		Usage();

	exit(0);
}

void
sendkeyshort(type)
	u_int type;
{
	struct sadb_msg *m_msg = (struct sadb_msg *)m_buf;

	m_len = sizeof(struct sadb_msg);

	m_msg->sadb_msg_version = PF_KEY_V2;
	m_msg->sadb_msg_type = type;
	m_msg->sadb_msg_errno = 0;
	m_msg->sadb_msg_satype = SADB_SATYPE_UNSPEC;
	m_msg->sadb_msg_len = _UNIT64(m_len);
	m_msg->sadb_msg_reserved = 0;
	m_msg->sadb_msg_reserved = 0;
	m_msg->sadb_msg_seq = 0;
	m_msg->sadb_msg_pid = getpid();

	sendkeymsg();

	return;
}

#if defined(PDEBUG)
static struct sockaddr_un saddr =
	{ sizeof(struct sockaddr_un), PF_UNIX, "/tmp/so.keydebug" };
#endif /* defined(PDEBUG) */

int
sendkeymsg()
{
	u_char rbuf[1024 * 32];	/* XXX: Enough ? Should I do MSG_PEEK ? */
	int so, len;

#if !defined(PDEBUG)
	if ((so = socket(PF_KEY, SOCK_RAW, PF_KEY_V2)) < 0) {
		perror("socket(PF_KEY)");
		goto end;
	}
    {
	struct timeval tv;
	tv.tv_sec = 1;
	tv.tv_usec = 0;
	if (setsockopt(so, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv)) < 0) {
		perror("setsockopt");
		goto end;
	}
    }
#else
	if ((so = socket(PF_UNIX, SOCK_STREAM, 0)) < 0) {
		perror("socket(PF_KEY)");
		goto end;
	}

	if (connect(so, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
		perror("connect");
		goto end;
	}
#endif /* defined(PDEBUG) */

	if (f_verbose)
		kdebug_sadb((struct sadb_msg *)m_buf);

	if ((len = send(so, m_buf, m_len, 0)) < 0) {
		perror("send");
		goto end;
	}

	if ((len = recv(so, rbuf, sizeof(rbuf), 0)) < 0) {
		perror("recv");
		goto end;
	}

	if (f_verbose)
		kdebug_sadb((struct sadb_msg *)rbuf);

	if (postproc(so, rbuf) < 0)
		goto end;

end:
	(void)close(so);
	return(0);
}

int
postproc(so, buf)
	int so;
	caddr_t buf;
{
	struct sadb_msg *msg = (struct sadb_msg *)buf;

	if (msg->sadb_msg_errno != 0) {
		u_char inf[80];
		char *errmsg = NULL;

		if (f_script)
			snprintf(inf, sizeof(inf), "The result of line %d: ", lineno);
		else
			inf[0] = '\0';

		switch (msg->sadb_msg_errno) {
		case ENOENT:
			switch (msg->sadb_msg_type) {
			case SADB_DELETE:
			case SADB_GET:
			case SADB_X_SPDDELETE:
				errmsg = "No entry";
				break;
			case SADB_DUMP:
				errmsg = "No SAD entries";
				break;
			case SADB_X_SPDDUMP:
				errmsg = "No SPD entries";
				break;
			}
			break;
		default:
			errmsg = strerror(msg->sadb_msg_errno);
		}
		printf("%s%s.\n", inf, errmsg);
		return(-1);
	}

	switch (msg->sadb_msg_type) {
	case SADB_GET:
		dump_msg(msg, SADB_DUMP);
		break;

	case SADB_DUMP:
		while (1) {
			dump_msg(msg, SADB_DUMP);
			if (msg->sadb_msg_seq == 0) break;
			msg = (struct sadb_msg *)((caddr_t)msg +
						_UNUNIT64(msg->sadb_msg_len));
			if (f_verbose)
				kdebug_sadb((struct sadb_msg *)msg);
		}
		break;

	case SADB_X_SPDDUMP:
		while (1) {
			dump_msg(msg, SADB_X_SPDDUMP);
			if (msg->sadb_msg_seq == 0) break;
			msg = (struct sadb_msg *)((caddr_t)msg +
						_UNUNIT64(msg->sadb_msg_len));
			if (f_verbose)
				kdebug_sadb((struct sadb_msg *)msg);
		}
		break;
	}

	return(0);
}

/*
 * Must to be re-written.
 */
caddr_t _str_satype[] = {
	"unspec",
	"unknown",
	"ah",
	"esp",
	"rsvp",
	"ospfv2",
	"ripv2",
	"mip",
};

caddr_t _str_upper[] = {
	"any", "icmp", "", "", "", "",
	"tcp", "", "", "", "", "", "", "", "", "", "",
	"udp",
};

caddr_t _str_base[] = {
	"new",
	"old",
};

caddr_t _str_state[] = {
	"larval",
	"mature",
	"dying",
	"dead",
};

caddr_t _str_alg_auth[] = {
	"none",
	"hmac-md5",
	"hmac-sha1",
	"md5",
	"sha",
	"null",
};

caddr_t _str_alg_enc[] = {
	"none",
	"des-cbc",
	"3des-cbc",
	"null",
	"blowfish-cbc",
	"cast128-cbc",
	"rc5-cbc",
};

caddr_t _str_sec_level[] = {
	"BYPASS ",
	"NONE ",
	"DEFAULT ",
	"AVAIL ",
	"USE ",
	"REQUIRE ",
	"UNIQUE ",
};

void
dump_msg(m, mode)
	struct sadb_msg *m;
	int mode;	/* SADB_DUMP or SADB_X_SPDDUMP */
{
	struct sadb_ext *ext;
	int tlen, extlen;

	u_int32_t d_spi;
	u_int d_replay, d_flags;
	u_int d_proto, d_fm, d_prefs, d_prefd, d_ports, d_portd, d_fmp;
	u_int d_key_enc_len, d_key_auth_len;
	caddr_t d_state, d_auth, d_enc;
	caddr_t d_src, d_dst, d_proxy;
	caddr_t d_key_auth, d_key_enc;
	u_char d_level[64];
	time_t d_cur_addtime, d_hard_addtime, d_soft_addtime;

	/* initialize */
	d_spi = 0;
	d_replay = d_flags = 0;
	d_proto = d_fm = d_prefs = d_prefd = d_ports = d_portd = d_fmp = 0;
	d_key_enc_len = d_key_auth_len = 0;
	d_state = d_auth = d_enc = NULL;
	d_src = d_dst = d_proxy = NULL;
	d_key_auth = d_key_enc = NULL;
	memset(d_level, 0, sizeof(d_level));
	d_cur_addtime = d_hard_addtime = d_soft_addtime = 0;

	tlen = _UNUNIT64(m->sadb_msg_len) - sizeof(struct sadb_msg);
	ext = (struct sadb_ext *)((caddr_t)m + sizeof(struct sadb_msg));

	while (tlen > 0) {
		switch (ext->sadb_ext_type) {
		case SADB_EXT_SA:
		    {
			struct sadb_sa *sa = (struct sadb_sa *)ext;

			d_spi = ntohl(sa->sadb_sa_spi);
			d_replay = sa->sadb_sa_replay;
			d_state = _str_state[sa->sadb_sa_state];
			d_auth = _str_alg_auth[sa->sadb_sa_auth];
			d_enc = _str_alg_enc[sa->sadb_sa_encrypt];
			d_flags = sa->sadb_sa_flags;
		    }
			break;

		case SADB_EXT_LIFETIME_CURRENT:
		    {
			struct sadb_lifetime *lft = (struct sadb_lifetime *)ext;

			d_cur_addtime = lft->sadb_lifetime_addtime;
		    }
			break;

		case SADB_EXT_LIFETIME_HARD:
		    {
			struct sadb_lifetime *lft = (struct sadb_lifetime *)ext;

			d_hard_addtime = lft->sadb_lifetime_addtime;
		    }
			break;

		case SADB_EXT_LIFETIME_SOFT:
		    {
			struct sadb_lifetime *lft = (struct sadb_lifetime *)ext;

			d_soft_addtime = lft->sadb_lifetime_addtime;
		    }
			break;

		case SADB_EXT_ADDRESS_SRC:
		    {
			struct sadb_address *ad = (struct sadb_address *)ext;
			struct sockaddr *sa;

			d_proto = ad->sadb_address_proto;
			d_prefs = ad->sadb_address_prefixlen;
			sa = (struct sockaddr *)((caddr_t)ext + sizeof(*ad));
			d_fm = sa->sa_family;
			d_ports = _INPORTBYAF(sa);
			d_src = _INADDRBYAF(sa);
		    }
			break;

		case SADB_EXT_ADDRESS_DST:
		    {
			struct sadb_address *ad = (struct sadb_address *)ext;
			struct sockaddr *sa;

			d_proto = ad->sadb_address_proto;
			d_prefd = ad->sadb_address_prefixlen;
			sa = (struct sockaddr *)((caddr_t)ext + sizeof(*ad));
			d_fm = sa->sa_family;
			d_portd = _INPORTBYAF(sa);
			d_dst = _INADDRBYAF(sa);
		    }
			break;

		case SADB_EXT_ADDRESS_PROXY:
		    {
			struct sockaddr *sa;
			sa = (struct sockaddr *)((caddr_t)ext + sizeof(struct sadb_address));

			d_fmp = sa->sa_family;
			d_proxy = _INADDRBYAF(sa);
		    }
			break;

		case SADB_EXT_KEY_AUTH:
		    {
			struct sadb_key *key;

			key = (struct sadb_key *)ext;
			d_key_auth_len = key->sadb_key_bits / 8;
			d_key_auth = (caddr_t)key + sizeof(*key);
		    }
			break;

		case SADB_EXT_KEY_ENCRYPT:
		    {
			struct sadb_key *key;

			key = (struct sadb_key *)ext;
			d_key_enc_len = key->sadb_key_bits / 8;
			d_key_enc = (caddr_t)key + sizeof(*key);
		    }
			break;

		case SADB_EXT_IDENTITY_SRC:
		case SADB_EXT_IDENTITY_DST:
		case SADB_EXT_SENSITIVITY:
		case SADB_EXT_PROPOSAL:
		case SADB_EXT_SUPPORTED_AUTH:
		case SADB_EXT_SUPPORTED_ENCRYPT:
		case SADB_EXT_SPIRANGE:
			break;

		case SADB_X_EXT_POLICY:
		    {
			struct sadb_x_policy *pl = (struct sadb_x_policy *)ext;

			switch (pl->sadb_x_policy_type) {
			case SADB_X_PL_DISCARD:
				strcpy(d_level, "DISCARD");
				break;
			case SADB_X_PL_BYPASS:
				strcpy(d_level, "BYPASS");
				break;
			case SADB_X_PL_IPSEC:
				strcat(d_level,
				 _str_sec_level[pl->sadb_x_policy_esp_trans]);
				strcat(d_level,
				 _str_sec_level[pl->sadb_x_policy_ah_trans]);
				strcat(d_level,
				 _str_sec_level[pl->sadb_x_policy_esp_network]);
				strcat(d_level,
				 _str_sec_level[pl->sadb_x_policy_ah_network]);
				break;
			default:
				printf("dump_msg: invalid policy type %u\n",
					pl->sadb_x_policy_type);
				return;
				break;
			}
		    }
			break;

		default:
			printf("dump_msg: invalid extension type %u\n",
				ext->sadb_ext_type);
			return;
			break;
		}

		extlen = _UNUNIT64(ext->sadb_ext_len);
		tlen -= extlen;
		ext = (struct sadb_ext *)((caddr_t)ext + extlen);
	}

	printf("%s ", _str_addr(d_src, d_prefs, d_ports, d_fm));
	printf("%s ", _str_addr(d_dst, d_prefd, d_portd, d_fm));
	printf("%s ", _str_upper[d_proto]);

	if (mode == SADB_X_SPDDUMP) {
		/* dump SPD */
		printf("%s\n", d_level);
		printf("\tstate=%s\n", _str_state[m->sadb_msg_satype]); /* XXX */
		printf("\trefcnt: %d\n", m->sadb_msg_reserved); /* TEST */

		return;
	}

	printf("%s\n", _str_addr(d_proxy, 0, 0, d_fmp));

	printf("\t%s spi=0x%08x replay=%u flags=0x%08lx\n",
		_str_satype[m->sadb_msg_satype],
		d_spi,
		d_replay, d_flags);

	if (d_key_enc != 0) {
		printf("\tE: %s,", d_enc);
		kdebug_pdump(d_key_enc, d_key_enc_len, KDUMP_HEX);
	}

	if (d_key_auth != 0) {
		printf("\tA: %s,", d_auth);
		kdebug_pdump(d_key_auth, d_key_auth_len, KDUMP_HEX);
	}

	printf("\tstate=%s seq=%ld pid=%ld\n",
		d_state,
		(u_long)m->sadb_msg_seq,
		(u_long)m->sadb_msg_pid);

    {
	time_t tmp_time = time(0);

	printf("\tcreated: %s", _str_time(d_cur_addtime));
	printf("\tcurrent: %s\n", _str_time(tmp_time));
	printf("\tdiff: %lu(s)",
		d_cur_addtime == 0 ? 0 : (tmp_time - d_cur_addtime));
	printf("\thard: %lu(s)", d_hard_addtime);
	printf("\tsoft: %lu(s)\n", d_soft_addtime);
    }

    {	/* TEST */
	caddr_t x = (caddr_t)&m->sadb_msg_reserved;
	printf("\trefcnt: sab=%d sa=%d\n", x[0], x[1]);
    }

	return;
}

u_char *
_str_addr(addr, pref, port, af)
	caddr_t addr;
	u_int pref, port, af;
{
	void inet_ntop();
	static u_char buf[128];
	u_char pbuf[128];

	if (addr == 0) return("");

	inet_ntop(af, addr, pbuf, sizeof(pbuf));

	if (pref != 0)
		snprintf(buf, sizeof(buf), "%s/%u[%u]", pbuf, pref, port);
	else
		snprintf(buf, sizeof(buf), "%s", pbuf);

	return(buf);
}

u_char *
_str_time(t)
	time_t t;
{
	static u_char buf[128];

	if (t == 0) {
		int i = 0;
		for (;i < 20;) buf[i++] = ' ';
	} else {
		char *t0;
		t0 = ctime(&t);
		memcpy(buf, t0 + 4, 20);
	}

	buf[20] = '\0';

	return(buf);
}

