/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*	BSDI inet.c,v 2.8 1996/11/12 20:08:05 jch Exp	*/
/*
 * Copyright (c) 1983, 1988, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char sccsid[] = "@(#)inet.c	8.4 (Berkeley) 4/20/94";
#endif /* not lint */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <sys/mbuf.h>
#include <sys/protosw.h>
#include <sys/sysctl.h>

#include <net/route.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>
#include <netinet6/in6_systm.h>
#include <netinet6/in6_pcb.h>
#include <netinet6/ip6_var.h>
#include <netinet6/tcp6.h>
#include <netinet6/tcp6_seq.h>
#define TCP6STATES
#include <netinet6/tcp6_fsm.h>
#include <netinet6/tcp6_timer.h>
#include <netinet6/tcp6_var.h>
#include <netinet6/tcp6_debug.h>
#include <netinet6/udp6.h>
#include <netinet6/udp6_var.h>
#ifdef IPSEC
#include <netinet6/ipsec.h>
#endif

#if 0
#include <arpa/inet.h>
#endif
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "netstat.h"

struct	in6pcb in6pcb;
struct	tcp6cb tcp6cb;
struct	socket sockb;

char	*inet6name __P((struct in6_addr *));
void	inet6print __P((struct in6_addr *, int, char *));

static char ntop_buf[INET6_ADDRSTRLEN];

/*
 * Print a summary of connections related to an Internet
 * protocol.  For TCP, also give state of connection.
 * Listening processes (aflag) are suppressed unless the
 * -a (all) flag is specified.
 */
void
ip6protopr(off, name)
	u_long off;
	char *name;
{
	struct in6pcb cb;
	struct in6pcb *prev, *next;
	int istcp;
	static int first = 1;

	if (off == 0)
		return;
	istcp = strcmp(name, "tcp6") == 0;
	kread(off, (char *)&cb, sizeof (struct in6pcb));
	in6pcb = cb;
	prev = (struct in6pcb *)off;
	if (in6pcb.in6p_next == (struct in6pcb *)off)
		return;
	while (in6pcb.in6p_next != (struct in6pcb *)off) {
		next = in6pcb.in6p_next;
		kread((u_long)next, (char *)&in6pcb, sizeof (in6pcb));
		if (in6pcb.in6p_prev != prev) {
			printf("???\n");
			break;
		}
#if 0
		if (!aflag &&
		  inet_lnaof(in6pcb.in6p_laddr) == INADDR_ANY) {
			prev = next;
			continue;
		}
#endif
		kread((u_long)in6pcb.in6p_socket, (char *)&sockb, sizeof (sockb));
		if (istcp) {
			kread((u_long)in6pcb.in6p_ppcb,
			    (char *)&tcp6cb, sizeof (tcp6cb));
		}
		if (first) {
			printf("Active Internet6 connections");
			if (aflag)
				printf(" (including servers)");
			putchar('\n');
			if (Aflag)
				printf("%-8.8s ", "PCB");
			printf(Aflag ?
				"%-5.5s %-6.6s %-6.6s  %-18.18s %-18.18s %s\n" :
				"%-5.5s %-6.6s %-6.6s  %-22.22s %-22.22s %s\n",
				"Proto", "Recv-Q", "Send-Q",
				"Local Address", "Foreign Address", "(state)");
			first = 0;
		}
		if (Aflag)
			if (istcp)
				printf("%8lx ", in6pcb.in6p_ppcb);
			else
				printf("%8lx ", next);
		printf("%-5.5s %6ld %6ld ", name, sockb.so_rcv.sb_cc,
		    sockb.so_snd.sb_cc);
		inet6print(&in6pcb.in6p_laddr, (int)in6pcb.in6p_lport, name);
		inet6print(&in6pcb.in6p_faddr, (int)in6pcb.in6p_fport, name);
		if (istcp) {
			if (tcp6cb.t_state < 0 || tcp6cb.t_state >= TCP6_NSTATES)
				printf(" %d", tcp6cb.t_state);
			else
				printf(" %s", tcp6states[tcp6cb.t_state]);
		}
		putchar('\n');
		prev = next;
	}
}

/*
 * Dump TCP statistics structure.
 */
void
tcp6_stats(off, name)
	u_long off;
	char *name;
{
	struct tcp6stat tcp6stat;
	int mib[] = { CTL_NET, PF_INET6, IPPROTO_TCP, TCP6CTL_STATS };

	if (skread(name, mib, sizeof(mib)/sizeof(*mib),
	    off, &tcp6stat, sizeof(tcp6stat)))
		return;

	printf ("%s:\n", name);

#define	p(f, m) if (tcp6stat.f || sflag <= 1) \
    printf(m, (u_quad_t)tcp6stat.f, PLURAL(tcp6stat.f))
#define	p_5(f, m) if (tcp6stat.f || sflag <= 1) \
    printf(m, (u_quad_t)tcp6stat.f)
#define	p2(f1, f2, m) if (tcp6stat.f1 || tcp6stat.f2 || sflag <= 1) \
    printf(m, (u_quad_t)tcp6stat.f1, PLURAL(tcp6stat.f1), \
        (u_quad_t)tcp6stat.f2, PLURAL(tcp6stat.f2))
#define	p2_5(f1, f2, m) if (tcp6stat.f1 || tcp6stat.f2 || sflag <= 1) \
    printf(m, (u_quad_t)tcp6stat.f1, PLURAL(tcp6stat.f1), \
        (u_quad_t)tcp6stat.f2)
#define	p3(f, m) if (tcp6stat.f || sflag <= 1) \
    printf(m, (u_quad_t)tcp6stat.f, PLURALES(tcp6stat.f))

	p(tcp6s_sndtotal, "\t%qu packet%s sent\n");
	p2(tcp6s_sndpack, tcp6s_sndbyte,
		"\t\t%qu data packet%s (%qu byte%s)\n");
	p2(tcp6s_sndrexmitpack, tcp6s_sndrexmitbyte,
		"\t\t%qu data packet%s (%qu byte%s) retransmitted\n");
	p(tcp6s_sndrexmitfast, "\t\t%qu fast retransmit%s\n");
	p2_5(tcp6s_sndacks, tcp6s_delack,
		"\t\t%qu ack-only packet%s (%qu delayed)\n");
	p(tcp6s_sndurg, "\t\t%qu URG only packet%s\n");
	p(tcp6s_sndprobe, "\t\t%qu window probe packet%s\n");
	p(tcp6s_sndwinup, "\t\t%qu window update packet%s\n");
	p(tcp6s_sndctrl, "\t\t%qu control packet%s\n");
	p(tcp6s_rcvtotal, "\t%qu packet%s received\n");
	p2(tcp6s_rcvackpack, tcp6s_rcvackbyte, "\t\t%qu ack%s (for %qu byte%s)\n");
	p(tcp6s_rcvdupack, "\t\t%qu duplicate ack%s\n");
	p(tcp6s_rcvacktoomuch, "\t\t%qu ack%s for unsent data\n");
	p2(tcp6s_rcvpack, tcp6s_rcvbyte,
		"\t\t%qu packet%s (%qu byte%s) received in-sequence\n");
	p2(tcp6s_rcvduppack, tcp6s_rcvdupbyte,
		"\t\t%qu completely duplicate packet%s (%qu byte%s)\n");
	p(tcp6s_pawsdrop, "\t\t%qu old duplicate packet%s\n");
	p2(tcp6s_rcvpartduppack, tcp6s_rcvpartdupbyte,
		"\t\t%qu packet%s with some dup. data (%qu byte%s duped)\n");
	p2(tcp6s_rcvoopack, tcp6s_rcvoobyte,
		"\t\t%qu out-of-order packet%s (%qu byte%s)\n");
	p2(tcp6s_rcvpackafterwin, tcp6s_rcvbyteafterwin,
		"\t\t%qu packet%s (%qu byte%s) of data after window\n");
	p(tcp6s_rcvwinprobe, "\t\t%qu window probe%s\n");
	p(tcp6s_rcvwinupd, "\t\t%qu window update packet%s\n");
	p(tcp6s_rcvafterclose, "\t\t%qu packet%s received after close\n");
	p(tcp6s_rcvbadsum, "\t\t%qu discarded for bad checksum%s\n");
	p(tcp6s_rcvbadoff, "\t\t%qu discarded for bad header offset field%s\n");
	p_5(tcp6s_rcvshort, "\t\t%qu discarded because packet too short\n");
	p(tcp6s_connattempt, "\t%qu connection request%s\n");
	p(tcp6s_accepts, "\t%qu connection accept%s\n");
	p(tcp6s_badsyn, "\t%qu bad connection attempt%s\n");
	p(tcp6s_connects, "\t%qu connection%s established (including accepts)\n");
	p2(tcp6s_closed, tcp6s_drops,
		"\t%qu connection%s closed (including %qu drop%s)\n");
	p(tcp6s_conndrops, "\t%qu embryonic connection%s dropped\n");
	p2(tcp6s_rttupdated, tcp6s_segstimed,
		"\t%qu segment%s updated rtt (of %qu attempt%s)\n");
	p(tcp6s_rexmttimeo, "\t%qu retransmit timeout%s\n");
	p(tcp6s_timeoutdrop, "\t\t%qu connection%s dropped by rexmit timeout\n");
	p(tcp6s_persisttimeo, "\t%qu persist timeout%s\n");
	p(tcp6s_persistdrop, "\t%qu connection%s timed out in persist\n");
	p(tcp6s_keeptimeo, "\t%qu keepalive timeout%s\n");
	p(tcp6s_keepprobe, "\t\t%qu keepalive probe%s sent\n");
	p(tcp6s_keepdrops, "\t\t%qu connection%s dropped by keepalive\n");
	p(tcp6s_droppedsyn, "\t\t%qu syn%s dropped due to queue or memory full\n");
	p(tcp6s_predack, "\t%qu correct ACK header prediction%s\n");
	p(tcp6s_preddat, "\t%qu correct data packet header prediction%s\n");
	p3(tcp6s_pcbcachemiss, "\t%qu PCB cache miss%s\n");
	p_5(tcp6s_sc_added, "\t%qu SYN cache entries added\n");
	p_5(tcp6s_sc_completed, "\t\t%qu completed\n");
	p_5(tcp6s_sc_aborted, "\t\t%qu aborted (no space to build PCB)\n");
	p_5(tcp6s_sc_timed_out, "\t\t%qu timed out\n");
	p_5(tcp6s_sc_overflowed, "\t\t%qu dropped due to overflow\n");
	p_5(tcp6s_sc_bucketoverflow, "\t\t%qu dropped due to bucket overflow\n");
	p_5(tcp6s_sc_reset, "\t\t%qu dropped due to RST\n");
	p_5(tcp6s_sc_unreach, "\t\t%qu dropped due to ICMP unreachable\n");
	p(tcp6s_sc_dupesyn, "\t%qu duplicate SYN%s received for entries already in the cache\n");
	p(tcp6s_sc_dropped, "\t%qu SYN%s dropped (no route or no space)\n");

#undef	p
#undef	p_5
#undef	p2
#undef	p2_5
#undef	p3
}

/*
 * Dump UDP statistics structure.
 */
void
udp6_stats(off, name)
	u_long off;
	char *name;
{
	struct udp6stat udp6stat;
	u_quad_t delivered;
	int mib[] = { CTL_NET, PF_INET6, IPPROTO_UDP, UDP6CTL_STATS };

	if (skread(name, mib, sizeof(mib)/sizeof(*mib),
	    off, &udp6stat, sizeof(udp6stat)))
		return;

	printf("%s:\n", name);
#define	p(f, m) if (udp6stat.f || sflag <= 1) \
    printf(m, (u_quad_t)udp6stat.f, PLURAL(udp6stat.f))
#define	p_5(f, m) if (udp6stat.f || sflag <= 1) \
    printf(m, (u_quad_t)udp6stat.f)
	p(udp6s_ipackets, "\t%qu datagram%s received\n");
	p_5(udp6s_hdrops, "\t%qu with incomplete header\n");
	p_5(udp6s_badlen, "\t%qu with bad data length field\n");
	p_5(udp6s_badsum, "\t%qu with bad checksum\n");
	p_5(udp6s_nosum, "\t%qu with no checksum\n");
	p_5(udp6s_noport, "\t%qu dropped due to no socket\n");
#if 0
	p(udp6s_noportbcast, "\t%qu broadcast/multicast datagram%s dropped due to no socket\n");
#endif
	p_5(udp6s_fullsock, "\t%qu dropped due to full socket buffers\n");
	delivered = udp6stat.udp6s_ipackets -
		    udp6stat.udp6s_hdrops -
		    udp6stat.udp6s_badlen -
		    udp6stat.udp6s_badsum -
		    udp6stat.udp6s_noport -
		    udp6stat.udp6s_fullsock;
	if (delivered || sflag <= 1)
		printf("\t%qu delivered\n", delivered);
	p(udp6s_opackets, "\t%qu datagram%s output\n");
#undef	p
#undef	p_5

}
/*
 * Dump IPSEC statistics structure.
 */
void
ipsec_stats(off, name)
	u_long off;
	char *name;
{
	struct ipsecstat ipsecstat;
	u_quad_t delivered;
	int mib[] = { CTL_NET, PF_INET, IPPROTO_ESP, IPSECCTL_STATS };

	if (strcmp(name, "ipsec") == 0)
		mib[1] = PF_INET;
	else
		mib[1] = PF_INET6;
	if (skread(name, mib, sizeof(mib)/sizeof(*mib),
	    off, &ipsecstat, sizeof(ipsecstat)))
		return;

	printf("%s:\n", name);
#define	p0(f, m) if (ipsecstat.f || sflag <= 1) \
    printf(m, (u_quad_t)ipsecstat.f)
#define	p(f, m) if (ipsecstat.f || sflag <= 1) \
    printf(m, (u_quad_t)ipsecstat.f, PLURAL(ipsecstat.f))
#define	p_es(f, m) if (ipsecstat.f || sflag <= 1) \
    printf(m, (u_quad_t)ipsecstat.f, PLURALES(ipsecstat.f))
	p_es(in_success, "\t%u inbound process%s succeeded\n");
	p0(in_polvio, "\t%u inbound process's security policy violation\n");
	p0(in_nosa, "\t%u inbound SA is unavailable\n");
	p0(in_inval, "\t%u inbound processing failed due to EINVAL\n");
	p0(in_badspi, "\t%u failed getting a SPI\n");
	p0(out_success, "\t%u succeeded outbound process\n");
	p0(out_polvio, "\t%u outbound process's security policy violation\n");
	p0(out_nosa, "\t%u outbound SA is unavailable\n");
	p_es(out_inval, "\t%u outbound process%s failed due to EINVAL\n");
	p(out_noroute, "\t%u packet%s without route\n");

#undef	p0
#undef	p
#undef	p_es
}

static	char *ip6nh[] = {
	"hop by hop",
	"ICMP",
	"IGMP",
	"#3",
	"IP",
	"#5",
	"TCP",
	"#7",
	"#8",
	"#9",
	"#10",
	"#11",
	"#12",
	"#13",
	"#14",
	"#15",
	"#16",
	"UDP",
	"#18",
	"#19",	
	"#20",
	"#21",
	"IDP",
	"#23",
	"#24",
	"#25",
	"#26",
	"#27",
	"#28",
	"TP",	
	"#30",
	"#31",
	"#32",
	"#33",
	"#34",
	"#35",
	"#36",
	"#37",
	"#38",
	"#39",	
	"#40",
	"IP6",
	"#42",
	"routing",
	"fragment",
	"#45",
	"#46",
	"#47",
	"#48",
	"#49",	
	"ESP",
	"AH",
	"#52",
	"#53",
	"#54",
	"#55",
	"#56",
	"#57",
	"ICMP6",
	"no next header",	
	"destination option",
	"#61",
	"#62",
	"#63",
	"#64",
	"#65",
	"#66",
	"#67",
	"#68",
	"#69",	
	"#70",
	"#71",
	"#72",
	"#73",
	"#74",
	"#75",
	"#76",
	"#77",
	"#78",
	"#79",	
	"ISOIP",
	"#81",
	"#82",
	"#83",
	"#84",
	"#85",
	"#86",
	"#87",
	"#88",
	"#89",	
	"#80",
	"#91",
	"#92",
	"#93",
	"#94",
	"#95",
	"#96",
	"Ethernet",
	"#98",
	"#99",	
	"#100",
	"#101",
	"#102",
	"#103",
	"#104",
	"#105",
	"#106",
	"#107",
	"#108",
	"#109",	
	"#110",
	"#111",
	"#112",
	"#113",
	"#114",
	"#115",
	"#116",
	"#117",
	"#118",
	"#119",	
	"#120",
	"#121",
	"#122",
	"#123",
	"#124",
	"#125",
	"#126",
	"#127",
	"#128",
	"#129",	
	"#130",
	"#131",
	"#132",
	"#133",
	"#134",
	"#135",
	"#136",
	"#137",
	"#138",
	"#139",	
	"#140",
	"#141",
	"#142",
	"#143",
	"#144",
	"#145",
	"#146",
	"#147",
	"#148",
	"#149",	
	"#150",
	"#151",
	"#152",
	"#153",
	"#154",
	"#155",
	"#156",
	"#157",
	"#158",
	"#159",	
	"#160",
	"#161",
	"#162",
	"#163",
	"#164",
	"#165",
	"#166",
	"#167",
	"#168",
	"#169",	
	"#170",
	"#171",
	"#172",
	"#173",
	"#174",
	"#175",
	"#176",
	"#177",
	"#178",
	"#179",	
	"#180",
	"#181",
	"#182",
	"#183",
	"#184",
	"#185",
	"#186",
	"#187",
	"#188",
	"#189",	
	"#180",
	"#191",
	"#192",
	"#193",
	"#194",
	"#195",
	"#196",
	"#197",
	"#198",
	"#199",	
	"#200",
	"#201",
	"#202",
	"#203",
	"#204",
	"#205",
	"#206",
	"#207",
	"#208",
	"#209",	
	"#210",
	"#211",
	"#212",
	"#213",
	"#214",
	"#215",
	"#216",
	"#217",
	"#218",
	"#219",	
	"#220",
	"#221",
	"#222",
	"#223",
	"#224",
	"#225",
	"#226",
	"#227",
	"#228",
	"#229",	
	"#230",
	"#231",
	"#232",
	"#233",
	"#234",
	"#235",
	"#236",
	"#237",
	"#238",
	"#239",	
	"#240",
	"#241",
	"#242",
	"#243",
	"#244",
	"#245",
	"#246",
	"#247",
	"#248",
	"#249",	
	"#250",
	"#251",
	"#252",
	"#253",
	"#254",
	"#255",
};

/*
 * Dump IP statistics structure.
 */
void
ip6_stats(off, name)
	u_long off;
	char *name;
{
	struct ip6stat ip6stat;
	int first, i;
	int mib[] = { CTL_NET, PF_INET6, IPPROTO_IP, IPV6CTL_STATS };

	if (skread(name, mib, sizeof(mib)/sizeof(*mib),
	    off, &ip6stat, sizeof(ip6stat)))
		return;

	printf("%s:\n", name);

#define	p(f, m) if (ip6stat.f || sflag <= 1) \
    printf(m, (u_quad_t)ip6stat.f, PLURAL(ip6stat.f))
#define	p_5(f, m) if (ip6stat.f || sflag <= 1) \
    printf(m, (u_quad_t)ip6stat.f)

	p(ip6s_total, "\t%qu total packet%s received\n");
#if 0
	p(ip6s_badsum, "\t%qu bad header checksum%s\n");
#endif
	p_5(ip6s_toosmall, "\t%qu with size smaller than minimum\n");
	p_5(ip6s_tooshort, "\t%qu with data size < data length\n");
#if 0
	p_5(ip6s_badhlen, "\t%qu with header length < data size\n");
	p_5(ip6s_badlen, "\t%qu with data length < header length\n");
#endif
	p_5(ip6s_badoptions, "\t%qu with bad options\n");
	p_5(ip6s_badvers, "\t%qu with incorrect version number\n");
#if 0
	p_5(ip6s_badsrcintf, "\t%qu with bad source interface\n");
#endif
	p(ip6s_fragments, "\t%qu fragment%s received\n");
	p(ip6s_fragdropped, "\t%qu fragment%s dropped (dup or out of space)\n");
	p(ip6s_fragtimeout, "\t%qu fragment%s dropped after timeout\n");
	p(ip6s_fragoverflow, "\t%qu fragmented packet%s dropped due to overflow\n");
	p(ip6s_reassembled, "\t%qu packet%s reassembled ok\n");
	p(ip6s_delivered, "\t%qu packet%s for this host\n");
#if 0
	p(ip6s_noproto, "\t%qu packet%s for unknown/unsupported protocol\n");
#endif
	p(ip6s_forward, "\t%qu packet%s forwarded\n");
	p(ip6s_cantforward, "\t%qu packet%s not forwardable\n");
	p(ip6s_redirectsent, "\t%qu redirect%s sent\n");
	p(ip6s_localout, "\t%qu packet%s sent from this host\n");
	p(ip6s_rawout, "\t%qu packet%s sent with fabricated ip header\n");
	p(ip6s_odropped, "\t%qu output packet%s dropped due to no bufs, etc.\n");
	p(ip6s_noroute, "\t%qu output packet%s discarded due to no route\n");
	p(ip6s_fragmented, "\t%qu output datagram%s fragmented\n");
	p(ip6s_ofragments, "\t%qu fragment%s created\n");
	p(ip6s_cantfrag, "\t%qu datagram%s that can't be fragmented\n");
	p(ip6s_badscope, "\t%qu packet%s that violated scope rules\n");
	p(ip6s_notmember, "\t%qu multicast packet%s which we don't join\n");
	for (first = 1, i = 0; i < 256; i++)
		if (ip6stat.ip6s_nxthist[i] != 0) {
			if (first) {
				printf("\tInput histogram:\n");
				first = 0;
			}
			printf("\t\t%s: %qu\n", ip6nh[i],
			       ip6stat.ip6s_nxthist[i]);
		}
	printf("\tMbuf statistics:\n");
	printf("\t\t%qu one mbuf\n", ip6stat.ip6s_m1);
	for (first = 1, i = 0; i < 32; i++) {
		char ifbuf[IFNAMSIZ];
		if (ip6stat.ip6s_m2m[i] != 0) {		
			if (first) {
				printf("\t\ttwo or more mbuf::\n");
				first = 0;
			}
			printf("\t\t\t%s = %qu\n",
			       if_indextoname(i, ifbuf),
			       ip6stat.ip6s_m2m[i]);
		}
	}
	printf("\t\t%qu one ext mbuf\n", ip6stat.ip6s_mext1);
	printf("\t\t%qu two or more ext mbuf\n", ip6stat.ip6s_mext2m);
	p(ip6s_exthdrtoolong, "\t%qu packet%s whose headers are not continuous\n");
	p(ip6s_nogif, "\t%qu tunneling packet%s that can't find gif\n");
	p(ip6s_toomanyhdr, "\t%qu packet%s discarded due to too may headers\n");
#undef p
}

static	char *icmp6names[] = {
	"#0",
	"unreach",
	"packet too big",
	"time exceed",
	"parameter problem",
	"#5",
	"#6",
	"#7",
	"#8",
	"#9",
	"#10",
	"#11",
	"#12",
	"#13",
	"#14",
	"#15",
	"#16",
	"#17",
	"#18",
	"#19",	
	"#20",
	"#21",
	"#22",
	"#23",
	"#24",
	"#25",
	"#26",
	"#27",
	"#28",
	"#29",	
	"#30",
	"#31",
	"#32",
	"#33",
	"#34",
	"#35",
	"#36",
	"#37",
	"#38",
	"#39",	
	"#40",
	"#41",
	"#42",
	"#43",
	"#44",
	"#45",
	"#46",
	"#47",
	"#48",
	"#49",	
	"#50",
	"#51",
	"#52",
	"#53",
	"#54",
	"#55",
	"#56",
	"#57",
	"#58",
	"#59",	
	"#60",
	"#61",
	"#62",
	"#63",
	"#64",
	"#65",
	"#66",
	"#67",
	"#68",
	"#69",	
	"#70",
	"#71",
	"#72",
#define GETSERVBYPORT6(port, proto, ret)\
{\
	if (strcmp((proto), "tcp6") == 0)\
		(ret) = getservbyport((int)(port), "tcp");\
	else if (strcmp((proto), "udp6") == 0)\
		(ret) = getservbyport((int)(port), "udp");\
	else\
		(ret) = getservbyport((int)(port), (proto));\
};
	"#73",
	"#74",
	"#75",
	"#76",
	"#77",
	"#78",
	"#79",	
	"#80",
	"#81",
	"#82",
	"#83",
	"#84",
	"#85",
	"#86",
	"#87",
	"#88",
	"#89",	
	"#80",
	"#91",
	"#92",
	"#93",
	"#94",
	"#95",
	"#96",
	"#97",
	"#98",
	"#99",	
	"#100",
	"#101",
	"#102",
	"#103",
	"#104",
	"#105",
	"#106",
	"#107",
	"#108",
	"#109",	
	"#110",
	"#111",
	"#112",
	"#113",
	"#114",
	"#115",
	"#116",
	"#117",
	"#118",
	"#119",	
	"#120",
	"#121",
	"#122",
	"#123",
	"#124",
	"#125",
	"#126",
	"#127",
	"echo",
	"echo reply",	
	"multicast listener query",
	"multicast listener report",
	"multicast listener done",
	"router solicitation",
	"router advertisment",
	"neighbor solicitation",
	"neighbor advertisment",
	"redirect",
	"router renumbering",
	"#139",
	"who are you",
	"my name is",
};

/*
 * Dump ICMP statistics.
 */
void
icmp6_stats(off, name)
	u_long off;
	char *name;
{
	struct icmp6stat icmp6stat;
	int i, first;
	int mib[] = { CTL_NET, PF_INET6, IPPROTO_ICMPV6, ICMPV6CTL_STATS };

	if (skread(name, mib, sizeof(mib)/sizeof(*mib),
	    off, &icmp6stat, sizeof (icmp6stat)))
		return;
	
	printf("%s:\n", name);

#define	p(f, m) if (icmp6stat.f || sflag <= 1) \
    printf(m, (u_quad_t)icmp6stat.f, PLURAL(icmp6stat.f))

	p(icp6s_error, "\t%qu call%s to icmp_error\n");
	p(icp6s_canterror,
	    "\t%qu error%s not generated because old message was icmp\n");
	p(icp6s_toofreq,
	    "\t%qu error%s not generated because rate limitation\n");
	for (first = 1, i = 0; i < ICMP6_MAXTYPE + 1; i++)
		if (icmp6stat.icp6s_outhist[i] != 0) {
			if (first) {
				printf("\tOutput histogram:\n");
				first = 0;
			}
			printf("\t\t%s: %qu\n", icmp6names[i],
				(u_quad_t)icmp6stat.icp6s_outhist[i]);
		}
	p(icp6s_badcode, "\t%qu message%s with bad code fields\n");
	p(icp6s_tooshort, "\t%qu message%s < minimum length\n");
	p(icp6s_checksum, "\t%qu bad checksum%s\n");
	p(icp6s_badlen, "\t%qu message%s with bad length\n");
	for (first = 1, i = 0; i < ICMP6_MAXTYPE + 1; i++)
		if (icmp6stat.icp6s_inhist[i] != 0) {
			if (first) {
				printf("\tInput histogram:\n");
				first = 0;
			}
			printf("\t\t%s: %qu\n", icmp6names[i],
				(u_quad_t)icmp6stat.icp6s_inhist[i]);
		}
	p(icp6s_reflect, "\t%qu message response%s generated\n");
#undef p
}

/*
 * Pretty print an Internet address (net address + port).
 * If the nflag was specified, use numbers instead of names.
 */
void
inet6print(in6, port, proto)
	struct in6_addr *in6;
	int port;
	char *proto;
{
	struct servent *sp = 0;
	char line[80], *cp;
	int width;

	sprintf(line, "%.*s.", (Aflag && !nflag) ? 12 : 16, inet6name(in6));
	cp = index(line, '\0');
	if (!nflag && port)
		GETSERVBYPORT6(port, proto, sp);
	if (sp || port == 0)
		sprintf(cp, "%.8s", sp ? sp->s_name : "*");
	else
		sprintf(cp, "%d", ntohs((u_short)port));
	width = Aflag ? 18 : 22;
	printf(" %-*.*s", width, width, line);
}

/*
 * Construct an Internet address representation.
 * If the nflag has been supplied, give
 * numeric value, otherwise try for symbolic name.
 */
char *
inet6name(in6p)
	struct in6_addr *in6p;
{
	char *cp;
	static char line[50];
	struct hostent *hp;
	struct netent *np;
	static char domain[MAXHOSTNAMELEN + 1];
	static int first = 1;

	if (first && !nflag) {
		first = 0;
		if (gethostname(domain, MAXHOSTNAMELEN) == 0 &&
		    (cp = index(domain, '.')))
			(void) strcpy(domain, cp + 1);
		else
			domain[0] = 0;
	}
	cp = 0;
	if (!nflag && IN6_IS_ADDR_UNSPECIFIED(in6p)) {
		hp = gethostbyaddr((char *)in6p, sizeof (*in6p), AF_INET);
		if (hp) {
			if ((cp = index(hp->h_name, '.')) &&
			    !strcmp(cp + 1, domain))
				*cp = 0;
			cp = hp->h_name;
		}
	}
	if (IN6_IS_ADDR_UNSPECIFIED(in6p))
		strcpy(line, "*");
	else if (cp)
		strcpy(line, cp);
	else {
		sprintf(line, "%s", inet_ntop(AF_INET6, (void *)in6p, ntop_buf,
			sizeof(ntop_buf)));
	}
	return (line);
}
